/*
    Project: TCTEMP
    Version: V1.22
    File   : na_main.c

    Copyright (c) 2006-2012 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "na.h"
#include "na_image.h"
#include "na_tc.h"
#include "tc.h"

#define SYSTEM_INFORMATION_SIZE 0x10000

typedef struct
{
    TC_ENTROPY_POOL ep;
} ENTROPY_CONTEXT, *PENTROPY_CONTEXT;

typedef struct
{
    /* the image context must be the first field due to alignment issues */

    IMAGE_CONTEXT   ic;

    BYTE            rgbSystemInformation[SYSTEM_INFORMATION_SIZE];

    PCWSTR          pcwszVolumeName;
    WCHAR           wszNtVolumeName[MAX_NT_PATH];

    TC_CONTEXT      tc;

    TC_HEADER       tch;
    TC_BUFFER       tcb;
    TC_PASSWORD     tcpw;

    BOOL            fVolumeIsSparseFile;
    ULONGLONG       ullVolumeLength;

    ENTROPY_CONTEXT ec;

} TCTEMP_CONTEXT, *PTCTEMP_CONTEXT;

typedef struct {
    ULONGLONG ullVolumeSize;
    DWORD     dwSectorSize;
} IMAGE_FILE_INFO, *PIMAGE_FILE_INFO;

typedef struct {
    LARGE_INTEGER liFileOffset;
    LARGE_INTEGER liBeyondFinalZero;
} FILE_ZERO_DATA_INFORMATION, *PFILE_ZERO_DATA_INFORMATION;

/*----------------------------------------------------------------------------*/
/* Entropy                                                                    */

#define ES_SYSTEM_INFO(iSI) 0
#define ES_BASIC_FILE_INFO  0
#define ES_TSC              0
#define ES_VIA_RNG          0

#define NaFillEntropyPool(pec, iEntropySource, pData, cbDataSize) \
  TC_FillEntropyPool(&pec->ep, pData, cbDataSize);

typedef struct
{
    FILE_BASIC_INFORMATION fbi;
    WCHAR                  wszNtFileName[MAX_NT_PATH];

} FILE_INFO_LOCALS, *PFILE_INFO_LOCALS;

static BOOL NAAPI NaFillEntropyPoolWithFileInfo(PENTROPY_CONTEXT pec,
  PCWSTR pwszFileName, PVOID pBuffer, SIZE_T cbBufferSize)
{
    PFILE_INFO_LOCALS pfil;
    HANDLE            hFile;
    NTSTATUS          ntStatus;
    BOOL              fResult;

    if (cbBufferSize < sizeof(FILE_INFO_LOCALS)) {
        return FALSE;
    }

    pfil = (PFILE_INFO_LOCALS)pBuffer;
    if (!NaGetNtFileName(pfil->wszNtFileName, countof(pfil->wszNtFileName),
      pwszFileName)) {
        return FALSE;
    }

    fResult = FALSE;
    if (NaOpenFile(&hFile, pfil->wszNtFileName, FILE_READ_ATTRIBUTES,
      FILE_SHARE_READ|FILE_SHARE_WRITE, 0)) {
        if (NaQueryFileInfo(hFile, &pfil->fbi, sizeof(pfil->fbi),
          FileBasicInformation)) {
            NaFillEntropyPool(pec, ES_BASIC_FILE_INFO, &pfil->fbi,
              sizeof(pfil->fbi));
            fResult = TRUE;
        }
        NaCloseFile(hFile);
    }

    return TRUE;
}

static void NAAPI NaFillEntropyPoolWithSystemInformation(PENTROPY_CONTEXT pec,
  PVOID pBuffer, ULONG ulBufferSize)
{
    static const int rgiSI[] = {
        2, 3, 5, 6, 8, 11, 16, 21, 23, 33, 36, 43, 45, 51, 55, 0
    };

    ULONG     ulReturnLength;
    const int *pi;

    for (pi = rgiSI; *pi; pi++) {
        if (NaGetSystemInformation(*pi, pBuffer, ulBufferSize,
          &ulReturnLength)) {
            NaFillEntropyPool(pec, ES_SYSTEM_INFO(*pi), pBuffer,
              ulReturnLength);
        }
    }
}

extern unsigned __int64 rdtsc(void);

static void NAAPI NaFillEntropyPoolWithTimeStampCounter(PENTROPY_CONTEXT pec)
{
    unsigned __int64 ull;

    ull = rdtsc();
    NaFillEntropyPool(pec, ES_TSC, &ull, sizeof(ull));
}

extern int get_rng_bytes(unsigned __int8 *pbBuffer, size_t cbBufferSize);

static void NAAPI NaFillEntropyPoolWithHardwareRNG(PENTROPY_CONTEXT pec,
  PVOID pBuffer, ULONG ulBufferSize)
{
    if (get_rng_bytes((unsigned __int8 *)pBuffer, ulBufferSize)) {
        NaFillEntropyPool(pec, ES_VIA_RNG, pBuffer, ulBufferSize);
    }
}

/*----------------------------------------------------------------------------*/
/* Volume header                                                              */

static BOOL NAAPI NaReadVolumeHeader(PTCTEMP_CONTEXT pCtx)
{
    HANDLE h;
    BOOL   fResult;

    fResult = FALSE;
    if (NaOpenFile(&h, pCtx->wszNtVolumeName, GENERIC_READ, FILE_SHARE_READ,
      FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {

        LARGE_INTEGER li;
        ULONG         ul;

        pCtx->fVolumeIsSparseFile = FALSE;
        if (NaGetFileAttributes(h, &ul)) {
            pCtx->fVolumeIsSparseFile = (ul & FILE_ATTRIBUTE_SPARSE_FILE) != 0;
        }
        pCtx->ullVolumeLength = 0;
        NaGetFileLength(h, &pCtx->ullVolumeLength);

        li.QuadPart = 0;
        if (NaReadFile(h, &pCtx->tch, sizeof(pCtx->tch), NULL, &li)) {
            fResult = TRUE;
        }
        NaCloseFile(h);
    }

    return fResult;
}

static BOOL NAAPI NaWriteVolumeHeader(PTCTEMP_CONTEXT pCtx)
{
    HANDLE h;
    BOOL   fResult;

    fResult = FALSE;
    if (NaOpenFile(&h, pCtx->wszNtVolumeName, GENERIC_WRITE, 0,
      FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
        LARGE_INTEGER li;
        li.QuadPart = 0;
        if (NaWriteFile(h, &pCtx->tch, sizeof(pCtx->tch), NULL, &li)) {
            fResult = TRUE;
        }
        NaCloseFile(h);
    }

    return fResult;
}

/*----------------------------------------------------------------------------*/
/* Get sector size and volume size from image header                          */

static BOOL NAAPI NaGetImageInfo(PTCTEMP_CONTEXT pCtx, PIMAGE_FILE_INFO pIfi)
{
    HANDLE h;
    BOOL   fResult;

    fResult = FALSE;
    memset(pIfi, '\0', sizeof(IMAGE_FILE_INFO));
    if (NaOpenFile(&h, pCtx->ic.wszNtImageFileName, GENERIC_READ,
      FILE_SHARE_READ, FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
        LARGE_INTEGER li;
        PIMAGE_HEADER pih;
        li.QuadPart = 0;
        pih = (PIMAGE_HEADER) pCtx->ic.rgbHeaderSector;
        if (NaReadFile(h, pih, 512, NULL, &li)) {
            pIfi->ullVolumeSize = pih->wSectorSize * pih->ullTotalSectors;
            pIfi->dwSectorSize = pih->wSectorSize;
            fResult = TRUE;
        }
        NaCloseFile(h);
    }

    return fResult;
}

/*----------------------------------------------------------------------------*/
/* Free allocated blocks of a dynamic TC volume (sparse file)                 */

static BOOL NAAPI NaResetSparseFile(PTCTEMP_CONTEXT pCtx, ULONG ulSectorSize)
{
    HANDLE h;
    BOOL   fResult;

    NaAssert(ulSectorSize >= 512);

    if (!pCtx->fVolumeIsSparseFile || pCtx->ullVolumeLength < ulSectorSize
      || ulSectorSize < sizeof(pCtx->tch)) {
        return FALSE;
    }

    fResult = FALSE;
    if (NaOpenFile(&h, pCtx->wszNtVolumeName, GENERIC_WRITE, 0,
      FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {

        FILE_ZERO_DATA_INFORMATION fzdi;
        fzdi.liFileOffset.QuadPart = ulSectorSize;
        fzdi.liBeyondFinalZero.QuadPart = pCtx->ullVolumeLength;

#if DEBUG
        NaPrintf(L"Freeing unused volume sectors...\n");
#endif
        if (NaDeviceIoControl(h, FSCTL_SET_ZERO_DATA, &fzdi, sizeof(fzdi),
          NULL, 0, NULL)) {
            fResult = TRUE;
        }

        NaCloseFile(h);
    }

    return fResult;
}

/*----------------------------------------------------------------------------*/
/* Stack                                                                      */

void NAAPI NaWipeStack(void)
{
    BYTE rgb[2048];

    RtlSecureZeroMemory(rgb, sizeof(rgb));
}

/*----------------------------------------------------------------------------*/
/* main                                                                       */

int _cdecl main(int argc, char *argv[])
{
    PTCTEMP_CONTEXT  pCtx;
    UNICODE_STRING   us;
    ANSI_STRING      as;
    unsigned         uCipher;
    unsigned         uHash;
    unsigned         uIterationCount;
    int              i;
    IMAGE_FILE_INFO  ifi;

#if DEBUG
    if (argc && stricmp(argv[argc-1], "dbg") == 0) {
        argc--;
        argv[argc] = NULL;
        DbgBreakPoint();
    }
#endif

    /* Initialization */
    if (!NaInitialize(&pCtx, sizeof(TCTEMP_CONTEXT), MAX_IMAGE_SECTOR_SIZE)) {
        return STATUS_NO_MEMORY;
    }

    TC_InitEntropyPool(&pCtx->ec.ep, 0);
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithSystemInformation(&pCtx->ec,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithHardwareRNG(&pCtx->ec,
      pCtx->rgbSystemInformation, 64);
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    /* Default settings */
    uCipher = TC_AES;
    uHash   = TC_SHA512;
    pCtx->pcwszVolumeName       = L"c:\\tctemp6.tc";
    pCtx->ic.chDriveLetter      = L'T';
    pCtx->ic.pcwszImageFileName = L"c:\\tctemp6.img";

    /* Command line */
    if (argc > 1) {
        switch(argv[1][0]) {
        case 'A':
            if (argv[1][1] == 'T' && argv[1][2] == 'S') {
                uCipher = TC_AES_TWOFISH_SERPENT;
            } else if (argv[1][1] == 'T') {
                uCipher = TC_AES_TWOFISH;
            } else {
                uCipher = TC_AES;
            }
            break;
        case 'S':
            if (argv[1][1] == 'T' && argv[1][2] == 'A') {
                uCipher = TC_SERPENT_TWOFISH_AES;
            } else if (argv[1][1] == 'A') {
                uCipher = TC_SERPENT_AES;
            } else {
                uCipher = TC_SERPENT;
            }
            break;
        case 'T':
            if (argv[1][1] == 'S') {
                uCipher = TC_TWOFISH_SERPENT;
            } else {
                uCipher = TC_TWOFISH;
            }
            break;

        }
    }
    if (argc > 2) {
        switch(argv[2][0]) {
        case 'h':
            uHash = TC_SHA512;
            break;
        case 'r':
            uHash = TC_RIPEMD160;
            break;
        case 'w':
            uHash = TC_WHIRLPOOL;
            break;
        }
    }
    uIterationCount = (uHash == TC_RIPEMD160) ? 2000 : 1000;
    if (argc > 3) {
        int iMul = atoi(argv[3]);
        if (iMul > 1 && iMul <= 2000000) {
            uIterationCount *= (unsigned)iMul;
        } else if (iMul < -(int)uIterationCount) {
            uIterationCount = (unsigned) -iMul;
        }
    }
    if (argc > 4) {
        CHAR ch = (CHAR)toupper(argv[4][0]);
        if (ch >= 'A' && ch <= 'Z') {
            pCtx->ic.chDriveLetter = ch;
        }
    }
    if (argc > 5) {
        RtlInitAnsiString(&as, argv[5]);
        as.MaximumLength = ++as.Length; /* include terminating NUL */
        if (NT_SUCCESS(RtlAnsiStringToUnicodeString(&us, &as, TRUE))) {
            pCtx->pcwszVolumeName = us.Buffer;
        }
    }
    if (argc > 6) {
        RtlInitAnsiString(&as, argv[6]);
        as.MaximumLength = ++as.Length; /* include terminating NUL */
        if (NT_SUCCESS(RtlAnsiStringToUnicodeString(&us, &as, TRUE))) {
            pCtx->ic.pcwszImageFileName = us.Buffer;
        }
    }

#if DEBUG
    for (i = 0; i < argc; i++) {
        PSTR p = (PSTR)argv[i];
        NaPrintf(L"%S ", p);
    }
    NaPrintf(L"\n\n");
#endif

    /* Initial processing */
    if (!NaGetNtFileName(pCtx->ic.wszNtImageFileName,
      countof(pCtx->ic.wszNtImageFileName), pCtx->ic.pcwszImageFileName)) {
        NaDbgPrintf(L"Cannot get image file name!");
        NaUninitialize();
        return STATUS_INTERNAL_ERROR;
    }
    if (!NaGetNtFileName(pCtx->wszNtVolumeName, countof(pCtx->wszNtVolumeName),
      pCtx->pcwszVolumeName)) {
        NaUninitialize();
        NaDbgPrintf(L"Cannot get volume name!");
        return STATUS_INTERNAL_ERROR;
    }
    if (!NaGetNtDriveName(pCtx->ic.wszNtDriveName,
      countof(pCtx->ic.wszNtDriveName), pCtx->ic.chDriveLetter)) {
        NaUninitialize();
        NaDbgPrintf(L"Cannot get drive letter!");
        return STATUS_INTERNAL_ERROR;
    }

    /* Collect entropy */
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    NaReadVolumeHeader(pCtx);

    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithFileInfo(&pCtx->ec, pCtx->ic.pcwszImageFileName,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithSystemInformation(&pCtx->ec,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    /* Create volume and initialize file system */
    NaGetImageInfo(pCtx, &ifi);
    if (TC_MakeRandomHeader(&pCtx->tch, &pCtx->tcpw,
      &pCtx->ec.ep, uHash, uCipher, uIterationCount, &pCtx->tcb,
      ifi.ullVolumeSize, ifi.dwSectorSize)) {
        if (NaWriteVolumeHeader(pCtx)) {
            NaResetSparseFile(pCtx, ifi.dwSectorSize);
            if (NaTcOpen(&pCtx->tc)) {
                if (NaTcMountVolume(&pCtx->tc, pCtx->ic.chDriveLetter,
                  pCtx->pcwszVolumeName, pCtx->tcpw.rgbPassword,
                  pCtx->tcpw.cbPasswordSize)) {
#if DEBUG
                    NaPrintf(L"Restoring file system...\n");
#endif
                    if (!NaRestoreFileSystem(&pCtx->ic)) {
                        NaTcDismountVolume(&pCtx->tc, pCtx->ic.chDriveLetter);
                    }
                }
                NaTcClose(&pCtx->tc);
            }
        }
    }

    RtlSecureZeroMemory(pCtx, sizeof(*pCtx));
    NaWipeStack();

    NaUninitialize();

    return STATUS_SUCCESS;
}
