/*
    Project: TCTEMP
    Version: V1.22
    File   : tc_crypt.c

    Copyright (c) 2006-2012 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tc_priv.h"

/*----------------------------------------------------------------------------*/
/* Functions                                                                  */

#define COMPARE_KEYS(x,y) if (\
  pKeys[x].ullKey[0] == pKeys[y].ullKey[0] && \
  pKeys[x].ullKey[1] == pKeys[y].ullKey[1] && \
  pKeys[x].ullKey[2] == pKeys[y].ullKey[2] && \
  pKeys[x].ullKey[3] == pKeys[y].ullKey[3]) return TRUE

BOOL __stdcall TC_KeyIsWeak(PTC_KEY pKey, unsigned uCipher)
{
    /* - assume three primary keys and three secondary keys */
    /* - check that they are all mutually different         */

    typedef struct {
        TC_UINT64 ullKey[4];
    } KEY256;
    KEY256 *pKeys = (KEY256 *)pKey->rgbKeys;
    int    i;

    for (i = 1; i < 6; i++) COMPARE_KEYS(0, i);
    for (i = 2; i < 6; i++) COMPARE_KEYS(1, i);
    for (i = 3; i < 6; i++) COMPARE_KEYS(2, i);
    for (i = 4; i < 6; i++) COMPARE_KEYS(3, i);
    for (i = 5; i < 6; i++) COMPARE_KEYS(4, i);

    return FALSE;
}

#define SETUP(idx, cipher, key, key_length, offset_key2) \
  if (cipher##_desc.setup(key, key_length, 0, &pCtx->sk[idx]) != CRYPT_OK || \
      cipher##_desc.setup(key+offset_key2, key_length, 0, &pCtx->sk2[idx]) != CRYPT_OK) \
        goto error; \
    pCtx->pdCipher[idx] = &cipher##_desc;

BOOL __stdcall TCE_Init(PTCE_CONTEXT pCtx, PTC_KEY pKey, unsigned uCipher)
{
    memset(pCtx, '\0', sizeof(*pCtx));

    switch(uCipher) {
#ifdef RIJNDAEL
    case TC_AES:
        SETUP(0, aes, pKey->rgbKeys, 32, 32);
        break;
#endif
#ifdef SERPENT
    case TC_SERPENT:
        SETUP(0, serpent, pKey->rgbKeys, 32, 32);
        break;
#endif
#ifdef TWOFISH
    case TC_TWOFISH:
        SETUP(0, twofish, pKey->rgbKeys, 32, 32);
        break;
#endif
#if defined(RIJNDAEL) && defined(TWOFISH)
    case TC_AES_TWOFISH:
        SETUP(0, twofish, pKey->rgbKeys,      32, 64);
        SETUP(1, aes,     pKey->rgbKeys + 32, 32, 64);
        break;
#endif
#if defined(RIJNDAEL) && defined(TWOFISH) && defined(SERPENT)
    case TC_AES_TWOFISH_SERPENT:
        SETUP(0, serpent, pKey->rgbKeys,      32, 96);
        SETUP(1, twofish, pKey->rgbKeys + 32, 32, 96);
        SETUP(2, aes,     pKey->rgbKeys + 64, 32, 96);
        break;
#endif
#if defined(SERPENT) && defined(RIJNDAEL)
    case TC_SERPENT_AES:
        SETUP(0, aes,     pKey->rgbKeys,      32, 64);
        SETUP(1, serpent, pKey->rgbKeys + 32, 32, 64);
        break;
#endif
#if defined(SERPENT) && defined(TWOFISH) && defined(RIJNDAEL)
    case TC_SERPENT_TWOFISH_AES:
        SETUP(0, aes,     pKey->rgbKeys,      32, 96);
        SETUP(1, twofish, pKey->rgbKeys + 32, 32, 96);
        SETUP(2, serpent, pKey->rgbKeys + 64, 32, 96);
        break;
#endif
#if defined(TWOFISH) && defined(SERPENT)
    case TC_TWOFISH_SERPENT:
        SETUP(0, serpent, pKey->rgbKeys,      32, 64);
        SETUP(1, twofish, pKey->rgbKeys + 32, 32, 64);
        break;
#endif
    default:
        return FALSE;
    }

    return TRUE;

error:
    TCE_Close(pCtx);
    return FALSE;
}

void __stdcall TCE_SetIndex(PTCE_CONTEXT pCtx,
  unsigned __int64 ullILo, unsigned __int64 ullIHi)
{
    pCtx->ullI   = ullILo;
    pCtx->ullIHi = ullIHi;
}

void __stdcall TCE_SetDataUnitNumber(PTCE_CONTEXT pCtx,
  unsigned __int64 ullDataUnitNumber)
{
    TCE_SetIndex(pCtx, ullDataUnitNumber, 0);
}

static void __stdcall GfMul128_2(unsigned __int8 *T)
{
    unsigned __int64 *pullLo = (unsigned __int64 *) T;
    unsigned __int64 *pullHi = &pullLo[1];

    if (*pullHi & CONST64(0x8000000000000000)) {
        *pullHi <<= 1;
        if (*pullLo & CONST64(0x8000000000000000)) {
            *pullHi |= 1;
        }
        *pullLo <<= 1;
        T[0] ^= 0x87;
    } else {
        *pullHi <<= 1;
        if (*pullLo & CONST64(0x8000000000000000)) {
            *pullHi |= 1;
        }
        *pullLo <<= 1;
    }
}

BOOL __stdcall TCE_Encrypt(PTCE_CONTEXT pCtx, void *pDest, size_t cbDestSize)
{
    unsigned __int8 *pbDest;
    unsigned __int8 *pbDestEnd;
    int             iEA;

    TC_ASSERT(pCtx);
    TC_ASSERT(pCtx->pdCipher[0]);
    TC_ASSERT(pDest);
    TC_ASSERT(cbDestSize != 0);
    TC_ASSERT(cbDestSize % 16 == 0);
    TC_ASSERT(pCtx->pdCipher[0]->block_length == 16);

    for (iEA = 0; iEA < 3; iEA++) {

        if (pCtx->pdCipher[iEA] == NULL) break;

        pbDest = (unsigned __int8 *)pDest;
        pbDestEnd = pbDest + cbDestSize;

        /* T = E_K2(I) */
        *(unsigned __int64 *)(pCtx->rgbT)   = pCtx->ullI;
        *(unsigned __int64 *)(pCtx->rgbT+8) = pCtx->ullIHi;
        if ((pCtx->pdCipher[iEA]->ecb_encrypt(pCtx->rgbT, pCtx->rgbT,
            &pCtx->sk2[iEA])) != CRYPT_OK) {
            return FALSE;
        }

        while (pbDest != pbDestEnd) {

            /* C = E_K(P xor T) xor T */
            Xor128(pbDest, pCtx->rgbT);
            if ((pCtx->pdCipher[iEA]->ecb_encrypt(pbDest, pbDest,
                &pCtx->sk[iEA])) != CRYPT_OK) {
                return FALSE;
            }
            Xor128(pbDest, pCtx->rgbT);

            /* T = T GfMul128 2 */
            GfMul128_2(pCtx->rgbT);

            pbDest += 16;
        }

    }

    return TRUE;
}

BOOL __stdcall TCE_Decrypt(PTCE_CONTEXT pCtx, void *pDest, size_t cbDestSize)
{
    unsigned __int8 *pbDest;
    unsigned __int8 *pbDestEnd;
    int             iEA;

    TC_ASSERT(pCtx);
    TC_ASSERT(pCtx->pdCipher[0]);
    TC_ASSERT(pDest);
    TC_ASSERT(cbDestSize != 0);
    TC_ASSERT(cbDestSize % 16 == 0);
    TC_ASSERT(pCtx->pdCipher[0]->block_length == 16);

    for (iEA = 3; iEA--; ) {

        if (pCtx->pdCipher[iEA] == NULL) continue;

        pbDest = (unsigned __int8 *)pDest;
        pbDestEnd = pbDest + cbDestSize;

        /* T = E_K2(I) */
        *(unsigned __int64 *)(pCtx->rgbT)   = pCtx->ullI;
        *(unsigned __int64 *)(pCtx->rgbT+8) = pCtx->ullIHi;
        if ((pCtx->pdCipher[iEA]->ecb_encrypt(pCtx->rgbT, pCtx->rgbT,
            &pCtx->sk2[iEA])) != CRYPT_OK) {
            return FALSE;
        }

        while (pbDest != pbDestEnd) {

            /* P = D_K(C xor T) xor T */
            Xor128(pbDest, pCtx->rgbT);
            if ((pCtx->pdCipher[iEA]->ecb_decrypt(pbDest, pbDest,
              &pCtx->sk[iEA])) != CRYPT_OK) {
                return FALSE;
            }
            Xor128(pbDest, pCtx->rgbT);

            /* T = T GfMul128 2 */
            GfMul128_2(pCtx->rgbT);

            pbDest += 16;
        }

    }

    return TRUE;
}

void __stdcall TCE_Close(PTCE_CONTEXT pCtx)
{
    if (pCtx->pdCipher[0]) {
        pCtx->pdCipher[0]->done(&pCtx->sk[0]);
        pCtx->pdCipher[0]->done(&pCtx->sk2[0]);
    }
    if (pCtx->pdCipher[1]) {
        pCtx->pdCipher[1]->done(&pCtx->sk[1]);
        pCtx->pdCipher[1]->done(&pCtx->sk2[1]);
    }
    if (pCtx->pdCipher[2]) {
        pCtx->pdCipher[2]->done(&pCtx->sk[2]);
        pCtx->pdCipher[2]->done(&pCtx->sk2[2]);
    }
    memset(pCtx, '\0', sizeof(*pCtx));
}
